searchData={"items":[{"type":"module","title":":zigler","doc":"Parse transform module for using Zigler with erlang.\n\nFor the canonical example, see:\nhttps://www.erlang.org/doc/man/erl_id_trans.html","ref":":zigler.html"},{"type":"module","title":"Prerequisites - :zigler","doc":"In order to use Zigler in an erlang project, you must have the Elixir\nruntime.   You may do this any way you wish, but Zigler recommends\nrebar_mix:\n\nhttps://github.com/Supersonido/rebar_mix\n\nThere are instructions on how to make sure Elixir is available at\ncompile time for your erlang project.","ref":":zigler.html#module-prerequisites"},{"type":"module","title":"Building a Zig Module - :zigler","doc":"General documentation on parse transforms is very light.  To use zigler as\na parse transform:\n\n```erlang\n-module(my_erlang_module).\n-compile({parse_transform, zigler}).\n-export([...]).\n\n-zig_code(\"\npub fn hello_world() [] const u8 {\n  return \"Hello, world!\";\n}\n\")\n\n-zig_opts([{otp_app, my_app}]).\n```\n\nThis creates the `hello_world/0` function in your\nmodule which returns the \"Hello, world!\" binary.\n\nfor options to be delivered in the `zig_opts` attribute, see the\n`Zig` module documentation.\n\nNote that the `...` for the `nifs` option is not representable in erlang AST.\nInstead, use the atom `auto`.\n\n>","ref":":zigler.html#module-building-a-zig-module"},{"type":"module","title":"Note {: .warning } - :zigler","doc":">\n> Erlang integration is highly experimental and the interface\n> may be changed in the future.","ref":":zigler.html#module-note-warning"},{"type":"function","title":":zigler.parse_transform/2","doc":"performs a parse transformation on the AST for an erlang module,\nconverting public functions in the","ref":":zigler.html#parse_transform/2"},{"type":"type","title":":zigler.foo/0","doc":"foobarbaz","ref":":zigler.html#t:foo/0"},{"type":"module","title":"Zig","doc":"Inline NIF support for [Zig](https://ziglang.org)","ref":"Zig.html"},{"type":"module","title":"Motivation - Zig","doc":"> Zig is a general-purpose programming language designed for robustness,\n> optimality, and maintainability.\n\nThe programming philosophy of Zig matches up nicely with the programming\nphilosophy of the BEAM VM and in particular its emphasis on simplicity and\nstructure should very appealing to the practitioners of Elixir.\n\nThe following features make Zig extremely amenable to inline language\nsupport in a BEAM language:\n\n- simplicity.  Zig's syntax is definable in a simple YACC document and\n  Zig takes a stance against making its featureset more complex (though\n  it may evolve somewhat en route to 1.0)\n- Composability.  Zig is unopinionated about how to go about memory\n  allocations.  Its allocator interface is very easily able to be backed\n  by the BEAM's, which means that you have access to generic memory\n  allocation *strategies* through its composable allocator scheme.\n- C integration.  It's very easy to design C-interop between Zig and C.\n  Zigler has been designed to make it easier to use Zigler to build\n  C libraries than to use C directly see [Easy C](#module-easy-c).","ref":"Zig.html#module-motivation"},{"type":"module","title":"Guides - Zig","doc":"Please consult the following guides for detailed topics:\n\n- [Using Nifs](01-nifs.html)\n- [Collection datatypes](02-collections.html)\n- [Allocator strategies](03-allocators.html)\n- [Nif options](04-nif_options.html)\n- [Resources](05-resources.html)\n- [C integration](06-c_integration.html)\n- [Concurrency strategies](07-concurrency.html)\n- [Global module options](08-module_options.html)\n- [Raw calling](09-raw_nifs.html)\n- [Module callbacks](10-callbacks.html)\n\n>","ref":"Zig.html#module-guides"},{"type":"module","title":"Zig version support {: .warning } - Zig","doc":">\n> although the large-scale archictecture of zigler is settled,\n> zigler features may break backwards compatibility until zig reaches\n> 1.0","ref":"Zig.html#module-zig-version-support-warning"},{"type":"module","title":"Nerves Support - Zig","doc":"Nerves is supported out of the box, and Zigler will be able to seamlessly\ndetect the cross-compilation information (os, architecture, runtime) and\nbuild correctly for that target.","ref":"Zig.html#module-nerves-support"},{"type":"module","title":"Basic NIFs - Zig","doc":"In the BEAM, you can define a NIF by consulting the following [document](\nhttps://erlang.org/doc/man/erl_nif.html) and implementing the appropriate\nshared object/DLL callbacks.  However, Zigler will take care of all of\nthis for you.\n\nSimply `use Zig` in your module, providing the otp_app name as an option.\n\nThen, use the `sigil_Z/2` macro and write inline zig code.  To present a \nfunction as a nif in your module, simply export it from your code namespace \nby making it a `pub` function in your zig code.\n\n#### Example\n\n```elixir\ndefmodule BasicModule do\n  use Zig, otp_app: :zigler\n\n  ~Z\"\"\"\n  pub fn add_one(number: i64) i64 {\n      return number + 1;\n  }\n  \"\"\"\nend\n\ntest \"basic module with nif\" do\n  assert 48 = BasicModule.add_one(47)\nend\n```\n\n> #### otp_app setting {: .info }\n>\n> You should replace `:zigler` in the following example with the name of\n> your own app.  If no such app exists (e.g. you are using livebook or \n> are in the terminal or escript), you can use `:zigler` as a fallback.\n\nZigler will *automatically* fill out the appropriate NIF C template, compile\nthe shared object, and bind it into the module pre-compilation. In the above\nexample, there will be a `BasiceModule.add_one/1` function call created.\n\nZigler will also make sure that your statically-typed Zig data are guarded\nwhen you marshal it from the dynamically-typed BEAM world.  However, you may\nonly pass in and return certain types.  As an escape hatch, you may use\nthe [`beam.term`](beam.html#term) type which is a wrapped\n[`ERL_NIF_TERM`](https://www.erlang.org/doc/man/erl_nif.html#ERL_NIF_TERM) type.\nSee [`erl_nif`](https://www.erlang.org/doc/man/erl_nif.html).\n\n```elixir\ntest \"argument error when types are mismatched\" do\n  assert_raise ArgumentError, fn -> BasicModule.add_one(\"not a number\") end\nend\n```","ref":"Zig.html#module-basic-nifs"},{"type":"module","title":"I don't want to use inline Zig - Zig","doc":"```zig\n\\\\ .noinline.zig\npub fn add_one(number: i64) i64 {\n    return number + 1;\n}\n```\n\n```elixir\ndefmodule NoInline do\n  use Zig, otp_app: :zigler, zig_code_path: \".noinline.zig\"\nend\n\ntest \"non-inline zig\" do\n  assert 48 = NoInline.add_one(47)\nend\n```","ref":"Zig.html#module-i-don-t-want-to-use-inline-zig"},{"type":"module","title":"Advanced usage: Unsupported erl_nif functions - Zig","doc":"the `beam` import does not comprehensively provide support for all functions\nin `erl_nif.h`.  If you need access to a function in `erl_nif.h` that isn't \nprovided by zigler, you would do it in the following fashion:\n\n- import `erl_nif` into your zig code, typically under the `e` namespace.\n- retrieve `beam.context.env` and use that as your ErlNifEnv pointer.\n- use `beam.term` for function return types, which is a struct with a single\n  field, `v`, of type `ERL_NIF_TERM`.\n\n#### Example\n\n```elixir\ndefmodule WithErlNif do\n  use Zig, otp_app: :zigler\n\n  ~Z\"\"\"\n  const e = @import(\"erl_nif\");\n  const beam = @import(\"beam\");\n\n  pub fn add_one(number: u64) beam.term {\n      return .{.v = e.enif_make_uint64(beam.context.env, number + 1)};\n  }\n  \"\"\"\nend\n\ntest \"raw erl_nif_function\" do\n  assert 48 = WithErlNif.add_one(47)\nend\n```\n\n> #### beam.context.env is a threadlocal {: .warning}\n>\n> `beam.context.env` is a threadlocal variable, and is not available when\n> calling functions using `raw` mode.  See [Raw](9-raw-mode.html) mode calling\n> for more information.","ref":"Zig.html#module-advanced-usage-unsupported-erl_nif-functions"},{"type":"module","title":"Advanced usage: Manual marshalling - Zig","doc":"If you need to marshal your own data, you may use the `beam.get` and \n`beam.make` functions to marshal data to and from the BEAM world.\n\n#### Example\n\n```elixir\ndefmodule ManualMarshalling do\n  use Zig, otp_app: :zigler, nifs: [add_one: [spec: false]]\n\n  @spec add_one(integer) :: integer\n\n  ~Z\"\"\"\n  const beam = @import(\"beam\");\n\n  pub fn add_one(val: beam.term) !beam.term {\n      const number = try beam.get(i64, val, .{});\n      return beam.make(number + 1, .{});\n  }\n  \"\"\"\nend\n\ntest \"manual marshalling\" do\n  assert 48 = ManualMarshalling.add_one(47)\nend\n```\n\nFor more details on [`get`](beam.html#get) and [`make`](beam.html#make)\nfunctions see the [`beam`](beam.html) documentation.\n\n> #### Manual Term marshalling {: .warning }\n>\n> If you don't use automatic marshalling, Zigler will not be able\n> to provide the following conveniences:\n>\n> - argument error details.  The zig code will raise a generic\n>   BEAM `ArgumentError` but it won't have specific details about\n>   what the expected type was and which argument was in error.\n>\n> - dialyzer type information for your function.  You will have\n>   to supply that type information outside `~Z` block, as shown\n>   in the example.","ref":"Zig.html#module-advanced-usage-manual-marshalling"},{"type":"module","title":"Importing external files - Zig","doc":"If you need to write zig code outside of the module, just place it in\nthe same directory as your module.\n\nYou may either call imported functions from the external file, or\nforward a function from the external file, either strategy will \nwork correctly.\n\n#### Example\n\n```zig\n\\\\ .extra_code.zig\npub fn add_one(number: u64) u64 {\n    return number + 1;\n}\n```\n\n```elixir\ndefmodule ExternalImport do\n  use Zig, otp_app: :zigler\n\n  ~Z\"\"\"\n  const extra_code = @import(\".extra_code.zig\");\n  pub fn add_one(number: u64) u64 {\n      return extra_code.add_one(number);\n  }\n\n  pub const forwarded_add_one = extra_code.add_one;\n  \"\"\"\nend\n\ntest \"external imports by calling\" do\n  assert 48 = ExternalImport.add_one(47)\nend\n\ntest \"external imports by forwarding\" do\n  assert 48 = ExternalImport.forwarded_add_one(47)\nend\n```","ref":"Zig.html#module-importing-external-files"},{"type":"module","title":"Advanced Usage: Custom source location - Zig","doc":"By default, Zigler places generated source code in the same directory\nas the module that uses Zigler, however, you may specify a different\ndirectory:\n\n```elixir\ndefmodule CustomSourceLocation do\n  use Zig, otp_app: :zigler, dir: \"test/.custom_location\"\n  \n  ~Z\"\"\"\n  pub fn add_one(number: u64) u64 {\n      return number + 1;\n  }\n  \"\"\"\nend\n\ntest \"custom_location is built\" do\n  assert File.dir?(\"test/custom_location\")\n  assert File.exists?(\"test/.custom_location/.Elixir.CustomSourceLocation.zig\")\nend\n```","ref":"Zig.html#module-advanced-usage-custom-source-location"},{"type":"module","title":"Advanced usage: change staging directory location - Zig","doc":"By default, zigler stages files in `/tmp/{modulename}` directory.  In some cases\nthis will cause user collisions and permissions errors when trying to build modules\non multitenant systems.  If you need to change the staging directory, set the\n`ZIGLER_STAGING_ROOT` environment variable to the desired directory.  The\nrecommended staging directory is `~/.cache/zigler`.  NB: In the future, this\nmay become the default staging directory.","ref":"Zig.html#module-advanced-usage-change-staging-directory-location"},{"type":"function","title":"Zig.code/1","doc":"retrieves the zig code from any given module that was compiled with zigler","ref":"Zig.html#code/1"},{"type":"function","title":"Zig.nif_name/2","doc":"outputs a String name for the module.\n\nnote that for filesystem use, you must supply the extension.  For internal (BEAM) use, the\nfilesystem extension will be inferred.  Therefore we provide two versions of this function.","ref":"Zig.html#nif_name/2"},{"type":"macro","title":"Zig.sigil_Z/2","doc":"declares a string block to be included in the module's .zig source file.","ref":"Zig.html#sigil_Z/2"},{"type":"macro","title":"Zig.sigil_z/2","doc":"like `sigil_Z/2`, but lets you interpolate values from the outside\nelixir context using string interpolation (the `#{value}` form)","ref":"Zig.html#sigil_z/2"},{"type":"function","title":"Zig.version/0","doc":"default version of zig supported by this version of zigler.\n\n>","ref":"Zig.html#version/0"},{"type":"function","title":"API warning {: .warning } - Zig.version/0","doc":">\n> this API may change in the future.","ref":"Zig.html#version/0-api-warning-warning"},{"type":"module","title":"Zig.Formatter","doc":"","ref":"Zig.Formatter.html"},{"type":"function","title":"Zig.Formatter.features/1","doc":"","ref":"Zig.Formatter.html#features/1"},{"type":"function","title":"Zig.Formatter.format/2","doc":"","ref":"Zig.Formatter.html#format/2"},{"type":"function","title":"Zig.Formatter.format_string/1","doc":"","ref":"Zig.Formatter.html#format_string/1"},{"type":"exception","title":"Zig.CompileError","doc":"","ref":"Zig.CompileError.html"},{"type":"function","title":"Zig.CompileError.message/1","doc":"","ref":"Zig.CompileError.html#message/1"},{"type":"function","title":"Zig.CompileError.resolve/2","doc":"","ref":"Zig.CompileError.html#resolve/2"},{"type":"exception","title":"Zig.Type.ParseError","doc":"","ref":"Zig.Type.ParseError.html"},{"type":"function","title":"Zig.Type.ParseError.message/1","doc":"","ref":"Zig.Type.ParseError.html#message/1"},{"type":"module","title":"beam","doc":"","ref":"beam.html"},{"type":null,"title":"beam.alloc_env/0","doc":"","ref":"beam.html#alloc_env"},{"type":null,"title":"beam.allocator/0","doc":"","ref":"beam.html#allocator"},{"type":null,"title":"beam.binary_to_slice/1","doc":"","ref":"beam.html#binary_to_slice"},{"type":null,"title":"beam.binary_to_term/2","doc":"","ref":"beam.html#binary_to_term"},{"type":null,"title":"beam.cleanup/2","doc":"","ref":"beam.html#cleanup"},{"type":null,"title":"beam.clear_env/2","doc":"","ref":"beam.html#clear_env"},{"type":null,"title":"beam.compare/2","doc":"","ref":"beam.html#compare"},{"type":null,"title":"beam.context/0","doc":"","ref":"beam.html#context"},{"type":null,"title":"beam.copy/2","doc":"","ref":"beam.html#copy"},{"type":null,"title":"beam.free_env/1","doc":"","ref":"beam.html#free_env"},{"type":null,"title":"beam.general_purpose_allocator/0","doc":"","ref":"beam.html#general_purpose_allocator"},{"type":null,"title":"beam.get/3","doc":"","ref":"beam.html#get"},{"type":null,"title":"beam.get_env/0","doc":"","ref":"beam.html#get_env"},{"type":null,"title":"beam.independent_context/1","doc":"","ref":"beam.html#independent_context"},{"type":null,"title":"beam.make/2","doc":"","ref":"beam.html#make"},{"type":null,"title":"beam.make_empty_list/1","doc":"","ref":"beam.html#make_empty_list"},{"type":null,"title":"beam.make_error_atom/1","doc":"","ref":"beam.html#make_error_atom"},{"type":null,"title":"beam.make_error_pair/2","doc":"","ref":"beam.html#make_error_pair"},{"type":null,"title":"beam.make_general_purpose_allocator_instance/0","doc":"","ref":"beam.html#make_general_purpose_allocator_instance"},{"type":null,"title":"beam.make_into_atom/2","doc":"","ref":"beam.html#make_into_atom"},{"type":null,"title":"beam.make_list_cell/3","doc":"","ref":"beam.html#make_list_cell"},{"type":null,"title":"beam.make_ref/1","doc":"","ref":"beam.html#make_ref"},{"type":null,"title":"beam.make_stacktrace/2","doc":"","ref":"beam.html#make_stacktrace"},{"type":null,"title":"beam.raise_elixir_exception/3","doc":"","ref":"beam.html#raise_elixir_exception"},{"type":null,"title":"beam.raise_exception/2","doc":"","ref":"beam.html#raise_exception"},{"type":null,"title":"beam.raise_with_error_return/3","doc":"","ref":"beam.html#raise_with_error_return"},{"type":null,"title":"beam.release_binary/1","doc":"","ref":"beam.html#release_binary"},{"type":null,"title":"beam.self/1","doc":"","ref":"beam.html#self"},{"type":null,"title":"beam.send/3","doc":"","ref":"beam.html#send"},{"type":null,"title":"beam.term_to_binary/2","doc":"","ref":"beam.html#term_to_binary"},{"type":null,"title":"beam.wide_alignment_allocator/0","doc":"","ref":"beam.html#wide_alignment_allocator"},{"type":null,"title":"beam.yield/0","doc":"","ref":"beam.html#yield"},{"type":"type","title":"beam.ClearEnvReturn/1","doc":"","ref":"beam.html#ClearEnvReturn"},{"type":"type","title":"beam.Compared/0","doc":"","ref":"beam.html#Compared"},{"type":"type","title":"beam.Context/0","doc":"","ref":"beam.html#Context"},{"type":"type","title":"beam.ContextMode/0","doc":"","ref":"beam.html#ContextMode"},{"type":"type","title":"beam.Payload/1","doc":"","ref":"beam.html#Payload"},{"type":"type","title":"beam.Resource/3","doc":"","ref":"beam.html#Resource"},{"type":"type","title":"beam.TermType/0","doc":"","ref":"beam.html#TermType"},{"type":"type","title":"beam.Thread/1","doc":"","ref":"beam.html#Thread"},{"type":"type","title":"beam.ThreadedCallbacks/1","doc":"","ref":"beam.html#ThreadedCallbacks"},{"type":"type","title":"beam.env/0","doc":"","ref":"beam.html#env"},{"type":"type","title":"beam.event/0","doc":"","ref":"beam.html#event"},{"type":"type","title":"beam.loader/0","doc":"","ref":"beam.html#loader"},{"type":"type","title":"beam.monitor/0","doc":"","ref":"beam.html#monitor"},{"type":"type","title":"beam.pid/0","doc":"","ref":"beam.html#pid"},{"type":"type","title":"beam.port/0","doc":"","ref":"beam.html#port"},{"type":"type","title":"beam.term/0","doc":"","ref":"beam.html#term"},{"type":"type","title":"beam.tid/0","doc":"","ref":"beam.html#tid"},{"type":"extras","title":"Zigler","doc":"# Zigler\n\nLibrary test status:\n\n- ![](https://github.com/ityonemo/zigler/workflows/Elixir%20CI%20-%20Linux/badge.svg)\n- ![](https://github.com/ityonemo/zigler/workflows/Elixir%20CI%20-%20MacOS/badge.svg)","ref":"readme.html"},{"type":"extras","title":"Installation: Elixir - Zigler","doc":"","ref":"readme.html#installation-elixir"},{"type":"extras","title":"Obtaining Zig dependency - Zigler","doc":"Run `mix zig.get`","ref":"readme.html#obtaining-zig-dependency"},{"type":"extras","title":"Main Installation - Zigler","doc":"Zigler is [available in Hex](https://hex.pm/packages/zigler), and the package can be installed\nby adding `zigler` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:zigler, \"~> 0.13.0\", runtime: false}\n  ]\nend\n```","ref":"readme.html#main-installation"},{"type":"extras","title":"Installation: Erlang - Zigler","doc":"","ref":"readme.html#installation-erlang"},{"type":"extras","title":"Zig dependency - Zigler","doc":"TBD.\n\n`~/.cache/zigler/zig-linux- -0.11.1`","ref":"readme.html#zig-dependency"},{"type":"extras","title":"Main Installation - Zigler","doc":"Erlang is only supported via rebar3.  You must enable the rebar_mix plugin and \nadd zigler to your deps in rebar3.\n\nNote that erlang support is highly experimental.  Please submit issues if you \nhave difficulty.\n\n```erlang\n{plugins, [rebar_mix]}.\n\n{deps, [{zigler, \"0.13\"}]}.\n\n```","ref":"readme.html#main-installation"},{"type":"extras","title":"Documentation - Zigler","doc":"Docs can be found at [https://hexdocs.pm/zigler](https://hexdocs.pm/zigler).","ref":"readme.html#documentation"},{"type":"extras","title":"Currently supported platforms - Zigler","doc":"- Linux\n- FreeBSD (tested, but not subjected to CI)\n- MacOS\n\n- Nerves cross-compilation is supported out of the box.","ref":"readme.html#currently-supported-platforms"},{"type":"extras","title":"Zig Nifs made easy - Zigler","doc":"Wouldn't it be nice if you could make NIFs as easily as you can use the `asm`\nkeyword in C?\n\nThis is now possible, using the magic of Zig.\n\n```elixir\ndefmodule ExampleZig do\n  use Zig, otp_app: :zigler\n  ~Z\"\"\"\n  pub fn example_fun(value1: f64, value2: f64) bool {\n    return value1 > value2;\n  }\n  \"\"\"\nend\n\ntest \"example nifs\" do\n  assert ExampleZig.example_fun(0.8, -0.8)\n  refute ExampleZig.example_fun(0.1, 0.4)\nend\n```\n\nZigler will do automatic type marshalling between Elixir code and Zig code.\nIt will also convert trickier types into types you care about, for example:\n\n```elixir\ndefmodule ZigCollections do\n  use Zig, otp_app: :zigler\n  ~Z\"\"\"\n  pub fn string_count(string: []u8) i64 {\n    return @intCast(string.len);\n  }\n\n  pub fn list_sum(array: []f64) f64 {\n    var sum: f64 = 0.0;\n    for(array) | item | {\n      sum += item;\n    }\n    return sum;\n  }\n  \"\"\"\nend\n\ntest \"type marshalling\" do\n  assert 9 == ZigCollections.string_count(\"hello zig\")\n  assert 6.0 == ZigCollections.list_sum([1.0, 2.0, 3.0])\nend\n```\n\nMemory allocation with zigler is easy!  A standard BEAM allocator is provided for you,\nso any zig code you import will play nice with the BEAM.\n\n```elixir\ndefmodule Allocations do\n  use Zig, otp_app: :zigler\n  ~Z\"\"\"\n  const beam = @import(\"beam\");\n\n  pub fn double_atom(string: []u8) !beam.term {\n    var double_string = try beam.allocator.alloc(u8, string.len * 2);\n    defer beam.allocator.free(double_string);\n\n    for (string, 0..) | char, i | {\n      double_string[i] = char;\n      double_string[i + string.len] = char;\n    }\n\n    return beam.make_into_atom(double_string, .{});\n  }\n  \"\"\"\nend\n\ntest \"allocations\" do\n  assert :foofoo == Allocations.double_atom(\"foo\")\nend\n```\n\nIt is a goal for Zigler to make using *it* to bind C libraries easier\nthan using C to bind C libraries.  Here is an example:\n\n```elixir\nif {:unix, :linux} == :os.type() do\n  defmodule Blas do\n    use Zig,     \n      otp_app: :zigler,\n      c: [link_lib: {:system, \"blas\"}]\n  \n    ~Z\"\"\"\n    const beam = @import(\"beam\");\n    const blas = @cImport({\n        @cInclude(\"cblas.h\");\n    });\n\n    const BadArgs = error { badarg };\n  \n    pub fn blas_axpy(a: f64, x: []f64, y: []f64) ![]f64 {\n        if (x.len != y.len) return error.badarg;\n    \n        blas.cblas_daxpy(@intCast(x.len), a, x.ptr, 1, y.ptr, 1);\n    \n        return y;\n    }\n    \"\"\"\n  end\n  \n  test \"we can use a blas shared library\" do\n    # returns aX+Y\n    assert [11.0, 18.0] == Blas.blas_axpy(3.0, [2.0, 4.0], [5.0, 6.0])\n  end\nend\n```","ref":"readme.html#zig-nifs-made-easy"},{"type":"extras","title":"Documentation (Elixir-only) - Zigler","doc":"You can document nif functions, local functions, zig structs, variables, and types.\nIf you document a nif function, it will be a part of the module documentation, and\naccessible using the iex `h` method, etc.\n\nExample:\n\n```elixir\ndefmodule Documentation do\n  use Zig, otp_app: :zigler\n  ~Z\"\"\"\n  /// a zero-arity function which returns 47.\n  pub fn zero_arity() i64 {\n    return 47;\n  }\n  \"\"\"\nend\n```","ref":"readme.html#documentation-elixir-only"},{"type":"extras","title":"Formatting (Elixir-only) - Zigler","doc":"Zigler ships with a formatter.  To activate the formatter, adapt the following to your\n`.formatter.exs`:\n\n```elixir\n[\n  inputs: ~w[\n    {mix,.formatter,.credo}.exs\n    {config,lib,rel,test}/**/*.{ex,exs,zig}\n    installer/**/*.{ex,exs}\n  ],\n  plugins: [Zig.Formatter]\n]\n```","ref":"readme.html#formatting-elixir-only"},{"type":"extras","title":"Erlang support (highly experimental) - Zigler","doc":"Use of Zigler with erlang is possible using parse transforms.  You must obtain\nzigler using the `rebar3` and the `rebar_mix` plugin.  Modules with zigler\nnifs should inculde code into one or more `zig_code` attribute and pass \nzigler options (identical to the elixir options) into a `zig_opts` attribute.  \nZigler will then create appropriate functions matching the zig functions as\nit does with elixir.  Please not that some features (such as integers > 64 \nbits) are not currently supported in erlang, although nearly full feature parity\nis planned.\n\n```erlang\n-module(erlang_zigler_module).\n-compile({parse_transform, zigler}). \n-export([foo/1, foo/0]).\n\n-zig_code(\"\npub fn foo() i32 {\n    return 47;\n}\n\").\n\n-zig_opts([{otp_app, zigler}]).\n\nfoo(X) ->\n    47 + X.\n```","ref":"readme.html#erlang-support-highly-experimental"},{"type":"extras","title":"Zigler Principles - Zigler","doc":"1. Make being a good citizen of the BEAM easy.\n2. Use magic, but sparingly, only to prevent errors.\n3. Let the user see behind the curtain.\n4. Let the user opt out of magic.\n5. Magic shouldn't get in the way.","ref":"readme.html#zigler-principles"},{"type":"extras","title":"Using Nifs","doc":"# Using Nifs\n\nNifs are the entrypoint between your BEAM code and your C ABI code. Zigler provides semantics which\nare designed to make it easy to write safe NIF code with the safety, memory guarantees that Zig\nprovides.","ref":"01-nifs.html"},{"type":"extras","title":"Preamble - Using Nifs","doc":"Near the top of your module you should use the `use Zig` directive. This will activate Zigler to\nseek zig code blocks and convert them to functions. You must provide the `otp_app` option, which\nenables Zigler to find a directory to place compilation artifacts (such as libraries) so that they\ncan be shipped with releases. By default, this will be `/priv/lib`.\n\n>","ref":"01-nifs.html#preamble"},{"type":"extras","title":"compilation artifacts {: .warning} - Using Nifs","doc":">\n> defaults for compilation artifacts may change in future versions of zigler.\n\nNote that we can ship zig code in *any* module in-place, so they will live alongside other functions\nor even other macro alterations you make to the module. In this case, we'll build our code into a\nmodule that will be tested with ExUnit.\n\n```elixir\ndefmodule NifGuideTest do\n  use Zig, otp_app: :zigler\n  use ExUnit.Case, async: true\n```","ref":"01-nifs.html#compilation-artifacts-warning"},{"type":"extras","title":"Basic function writing - Using Nifs","doc":"Once zigler has been activated for a module, write `~Z` code anywhere and this code will be\nassembled into a zig file that will be compiled into the nif artifact. \n\nThen write your desired zig function Zigler will also mount functions with the *same name* as the\nfunction in the body of the module.","ref":"01-nifs.html#basic-function-writing"},{"type":"extras","title":"Example: simple scalar values - Using Nifs","doc":"```elixir\n~Z\"\"\"\npub fn add_one(input: i32) i32 {\n    return input + 1;\n}\n\"\"\"\n\ntest \"add one\" do\n  assert 48 == add_one(47)\nend\n```\n\nNote that Zigler will automatically marshal input and output values across the nif boundary. The\nfollowing scalar types are accepted by zigler:\n\n- signed integer (`i0`..`i65535`), including non-power-of-two values\n- unsigned integer (`u0`..`u65535`), including non-power-of-two values\n- `usize`, `isize`, architecture-dependent size (roughly `size_t` and `ssize_t` in C)\n- `c_char`, `c_short`, `c_ushort`, `c_int`, `c_uint`, `c_ulong`, `c_longlong`, `c_ulonglong`, which\n  are architecture-dependent integer sizes mostly used for c interop.\n- floats `f16`, `f32`, and `f64`.\n- `bool` (use the atoms `true` or `false` exclusively)\n\n> #### Floating point datatypes {: .info}\n>\n> Floating point datatypes can take the atoms `:infinity`, `:neg_infinity`, and `:NaN`.\n\n> #### Boolean datatypes {: .info}\n>\n> You must pass boolean datatypes `true` or `false` atoms.\n\nZigler can also marshal list parameters into array, and array-like datatypes:","ref":"01-nifs.html#example-simple-scalar-values"},{"type":"extras","title":"Example: Array-like datatypes - Using Nifs","doc":"```elixir\n~Z\"\"\"\npub fn sum(input: []f32) f32 {\n    var total: f32 = 0.0;\n    for (input) | value | { total += value; }\n    return total;\n}\n\"\"\"\n\ntest \"sum\" do\n  assert 6.0 == sum([1.0, 2.0, 3.0])\nend\n```\n\nThe following array-like datatypes are allowed for parameters:\n\n- arrays `[3]T` (for example). Note the length is compile-time known.\n- slices `[]T`\n- pointers to arrays `*[3]T` (for example).\n- multipointers `[*]T`\n- sentinel-terminated versions of all of the above.\n- cpointers `[*c]T`","ref":"01-nifs.html#example-array-like-datatypes"},{"type":"extras","title":"Example: Array-like datatypes as binaries - Using Nifs","doc":"For all scalar child types, Array-like datatypes may be passed as binaries, thus the following code\nworks with no alteration:\n\n```elixir\ntest \"sum, with binary input\" do\n  assert 6.0 == sum(<<1.0 :: float-size(32)-native, 2.0 :: float-size(32)-native, 3.0 :: float-size(32)-native>>)\nend\n```\n\nThis also results in a natural interface for treating BEAM binaries as `u8` arrays or slices.","ref":"01-nifs.html#example-array-like-datatypes-as-binaries"},{"type":"extras","title":"Example: Marshalling errors - Using Nifs","doc":"Zigler will generate code that protects you from sending incompatible datatypes to the desired\nfunction:\n\n```elixir\ntest \"marshalling error\" do\n  assert_raise ArgumentError, \"\"\"\n  errors were found at the given arguments:\n\n    * 1st argument: \n  \n       expected: list(float | :infinity | :neg_infinity | :NaN) | <<_::_*32>> (for `[]f32`)\n       got: `:atom`\n  \"\"\", fn ->\n    sum(:atom)\n  end\nend\n```\n\nFor more on marshalling collection datatypes, see [`collections`](collections.html).","ref":"01-nifs.html#example-marshalling-errors"},{"type":"extras","title":"Marshalling types manually - Using Nifs","doc":"You may also manually marshal types into and out of the beam by using the\n[`beam.term`](beam.html#term) datatype. To do so, you must first import the [`beam`](beam.html)\npackage. The `beam.term` type is an opaque, wrapped datatype that ensures safe manipulation of terms\nas a token in your zig code.\n\n```elixir\n~Z\"\"\"\nconst beam = @import(\"beam\");\n\npub fn manual_addone(value_term: beam.term) !beam.term {\n    const value = try beam.get(i32, value_term, .{});\n    return beam.make(value + 1, .{});\n}\n\"\"\"\n\ntest \"manual marshalling\" do\n  assert 48 == manual_addone(47)\nend\n```","ref":"01-nifs.html#marshalling-types-manually"},{"type":"extras","title":"Send - Using Nifs","doc":"An important mechanism for exporting values out of a NIF is to send it to a process\nID using the `send` function.  Zigler provides an advanced `beam.send` function which\nwill perform this operation for you.\n\n```elixir\n~Z\"\"\"\npub fn test_send(pid: beam.pid) !void {\n    try beam.send(pid, .{.ok, 47}, .{});\n}\n\"\"\"\n\ntest \"sending\" do\n  test_send(self())\n  assert_receive {:ok, 47}\nend\n```\n\nNote that the second argument of send and the options argument are equivalent to a `beam.make`\nfunction call.","ref":"01-nifs.html#send"},{"type":"extras","title":"Optional values - Using Nifs","doc":"You may use optional values as both input and output terms.  Note that the empty optional\nvalue in elixir is `nil` and the empty optional value in zig is `null`.\n\n```elixir\n~Z\"\"\"\npub fn optional(number: ?u32) ?u32 {\n    if (number) | n | {\n        if (n == 42) return null;\n        return n;\n    } else {\n        return 47;\n    }\n}\n\"\"\"\n\ntest \"optionals\" do\n  assert is_nil(optional(42))\n  assert 47 = optional(nil)\n  assert 10 = optional(10)\nend\n```","ref":"01-nifs.html#optional-values"},{"type":"extras","title":"Error Returns - Using Nifs","doc":"You may use functions which return an error, in which case an `ErlangException` will be\nthrown with the value being the atom representing the error.\n\n```elixir\n~Z\"\"\"\nconst OopsError = error { oops };\n\npub fn erroring() !void {\n    return error.oops;\n}\n\"\"\"\n\ntest \"erroring\" do\n  assert_raise ErlangError, \"Erlang error: :oops\", fn -> erroring() end\nend\n```","ref":"01-nifs.html#error-returns"},{"type":"extras","title":"A few notes on the above code. - Using Nifs","doc":"- to marshal a value out of a `beam.term` and into a zig static type, use [`beam.get`](beam.html#get).\n  This is a failable function, and in this case we hoist the failure in the function return.\n- to marshal a value into a `beam.term` from a zig static type, use [`beam.make`](beam.html#make).\n  This function *does not fail*.\n- for more information on the final options parameter of `get` and `make`, see their respective\n  documentation.\n- zigler will translate the hoisted marshalling failures into detailed BEAM exceptions of type `ArgumentError`.","ref":"01-nifs.html#a-few-notes-on-the-above-code"},{"type":"extras","title":"Collection Datatypes","doc":"# Collection Datatypes","ref":"02-collections.html"},{"type":"extras","title":"Returning array-like datatypes - Collection Datatypes","doc":"For array-like datatypes, we saw in [Using Nifs](1-nifs.html#example-array-like-datatypes) how they\ncan take both lists or binaries as inputs. However, when returning statically-typed data from zig, a\nchoice needs to be made as to whether to return lists or binaries.","ref":"02-collections.html#returning-array-like-datatypes"},{"type":"extras","title":"Arrays - Collection Datatypes","doc":"Arrays are the simplest array-like datatype to return.\n\n```elixir\n~Z\"\"\"\npub fn return_array(input: f32) [3]f32 {\n  var result: [3]f32 = undefined;\n\n  // set each item in the array:\n  for (&result, 0..) |*item, index| {\n    item.* = input + @as(f32, @floatFromInt(index));\n  }\n  return result;\n}\n\"\"\"\n\ntest \"returning an array\" do\n  assert [47.0, 48.0, 49.0] == return_array(47.0)\nend\n```","ref":"02-collections.html#arrays"},{"type":"extras","title":"Slices - Collection Datatypes","doc":"Slices can also be returned. Note that in many cases, returning a slice might need an [allocation strategy](3-allocators.html)\n\n```elixir\n~Z\"\"\"\npub fn return_slice(input: []f32) []f32 {\n  // set each item in the array:\n  for (input, 0..) |*item, index| {\n    item.* = item.* + @as(f32, @floatFromInt(index));\n  }\n  return input;\n}\n\"\"\"\n\ntest \"returning a slice\" do\n  assert [47.0, 48.0, 49.0] == return_slice([47.0, 47.0, 47.0])\nend\n```\n\n>","ref":"02-collections.html#slices"},{"type":"extras","title":"u8 array-likes output as binary {: .info} - Collection Datatypes","doc":">\n> u8 array-like datatypes are marshalled into binary by default instead of list.\n\n```elixir\n~Z\"\"\"\npub fn return_u8_array() [3]u8 {\n    const result: [3]u8 = .{97, 98, 99};\n    return result;\n}\n\"\"\"\n\ntest \"u8 datatypes are returned as binary\" do\n  assert \"abc\" == return_u8_array()\nend \n```","ref":"02-collections.html#u8-array-likes-output-as-binary-info"},{"type":"extras","title":"Selecting output type - Collection Datatypes","doc":"It's also possible to return these collections as binaries, however, in order to do so you will have\nto marshal manually. For datatypes that are more than 1 byte, be aware that the endianness of the\nresulting data is `native`.\n\n```elixir\n~Z\"\"\"\nconst beam = @import(\"beam\");\n\npub fn return_slice_binary(input: []f32) beam.term {\n  // set each item in the array:\n  for (input, 0..) |*item, index| {\n    item.* = item.* + @as(f32, @floatFromInt(index));\n  }\n  return beam.make(input, .{.as = .binary});\n}\n\"\"\"\n\ntest \"returning a slice as binary\" do\n  assert <<47.0 :: float-size(32)-native, \n           48.0 :: float-size(32)-native,\n           49.0 :: float-size(32)-native>> == return_slice_binary([47.0, 47.0, 47.0])\nend\n```\n\nConversely, for u8 array-like datatypes, selecting `.as = .list` will result in outputting a list.\n\nYou can also automatically marshal as binary by using [nif options](4-nif_options.html#binary-output)","ref":"02-collections.html#selecting-output-type"},{"type":"extras","title":"Full list of qualified array-like return types - Collection Datatypes","doc":"The central challenge of exporting array-like data to the BEAM is that length information may not be\nknown. In the case of arrays and slices, length is either comptime or runtime known. For other data\ntypes, the scope of datatypes accepted must be limited:\n\n- array of any type (`[_]T`)\n- single pointers to an array (`*[_]T`)\n- slices (`[]T`)\n- sentinel-terminated forms of the above (`[_:0]T` or `[:0]T`)\n- sentinel-terminated many-item-pointer (`[*:0]T`)\n- cpointer to u8 (`[*c]u8`). This will assume the cpointer is null-terminated, if it can't be\n  considered `[*:0]u8` the behaviour is undefined.\n- cpointer to a pointer (`[*c]?*T`). This will assume cpointer is null-terminated, if it can't be\n  considered `[*:null]?*T` then the behaviour is undefined.","ref":"02-collections.html#full-list-of-qualified-array-like-return-types"},{"type":"extras","title":"Passing and returning enums - Collection Datatypes","doc":"Enums are collections of integer values that are given special identifier status in the zig\nprogramming language. At compile-time, it's possible to get reflection on the string representation\nof those identifiers. Zigler thus is enabled to use enums as a representation of atoms coming from\nor going to the BEAM.","ref":"02-collections.html#passing-and-returning-enums"},{"type":"extras","title":"Enums as atoms - Collection Datatypes","doc":"Enums can be created by referring to them by the atom that corresponds to their value:\n\n```elixir\n~Z\"\"\"\nconst EnumType = enum(u8) {\n  foo, \n  bar = 47\n};\n\npub fn flip(input: EnumType) EnumType {\n  return switch (input) {\n    .foo => .bar,\n    .bar => .foo\n  };\n}\n\"\"\"\n\ntest \"flipping enums\" do\n  assert :bar = flip(:foo)\n  assert :foo = flip(:bar)\nend\n```","ref":"02-collections.html#enums-as-atoms"},{"type":"extras","title":"Enums as integers - Collection Datatypes","doc":"Functions taking an integer type can also by passed the associated integer value in the place of the\natom.\n\n```elixir\ntest \"enums passed as integer\" do\n  assert :foo = flip(47)\nend\n```","ref":"02-collections.html#enums-as-integers"},{"type":"extras","title":"Enum literals - Collection Datatypes","doc":"Enum literals can be converted to atoms using [`beam.make`](beam.html#make).\n\n```elixir\n~Z\"\"\"\npub fn make_literal() beam.term {\n  return beam.make(.some_new_literal, .{});\n}\n\"\"\"\n\ntest \"enum literals\" do\n  assert :some_new_literal = make_literal()\nend\n```\n\nThis is especially useful for emitting `:ok` or `:error` tuples.\n\n>","ref":"02-collections.html#enum-literals"},{"type":"extras","title":"error atom {: .info} - Collection Datatypes","doc":">\n> `error` is a reserved word in Zig, so to create error atom you must use builtin syntax; the error\n> enum literal is represented `.@\"error\"`.","ref":"02-collections.html#error-atom-info"},{"type":"extras","title":"Passing and returning structs - Collection Datatypes","doc":"","ref":"02-collections.html#passing-and-returning-structs"},{"type":"extras","title":"Structs are atom-keyed maps - Collection Datatypes","doc":"Most structs are interpreted as atom-keyed maps. Consider the following code:\n\n```elixir\n~Z\"\"\"\npub const Point2D = struct{ x: i32, y: i32 };\n\npub fn reflect(input: Point2D) Point2D {\n  return .{.x = input.y, .y = input.x};\n}\n\"\"\"\n\ntest \"structs\" do\n  assert %{x: 48, y: 47} == reflect(%{x: 47, y: 48})\nend\n```\n\n>","ref":"02-collections.html#structs-are-atom-keyed-maps"},{"type":"extras","title":"Structs in parameters and returns {: .info} - Collection Datatypes","doc":">\n> for a struct type to be used in parameters and returns, it must be exported as `pub` in the module\n> interface","ref":"02-collections.html#structs-in-parameters-and-returns-info"},{"type":"extras","title":"Anonymous structs can be returned - Collection Datatypes","doc":"It's possible to return anonymous structs as well, using [`beam.make`](beam.html#make).\n\n```elixir\n~Z\"\"\"\npub fn anonymous_struct() beam.term {\n    return beam.make(.{.foo = .bar}, .{});\n}\n\"\"\"\n\ntest \"anonymous structs\" do\n  assert %{foo: :bar} == anonymous_struct()\nend\n```","ref":"02-collections.html#anonymous-structs-can-be-returned"},{"type":"extras","title":"Zig tuples are structs. - Collection Datatypes","doc":"Tuples in zig are structs (with hidden integer-valued keys). Zigler takes advantage of this and\nallows you to construct BEAM tuples using zig tuples, when passed to [`beam.make`](beam.html#make).\n\n```elixir\n~Z\"\"\"\npub fn tuple() beam.term {\n    return beam.make(.{.ok, 47}, .{});\n}\n\"\"\"\n\ntest \"tuples\" do\n  assert {:ok, 47} == tuple()\nend\n```","ref":"02-collections.html#zig-tuples-are-structs"},{"type":"extras","title":"Packed or Extern structs. - Collection Datatypes","doc":"Packed or Extern structs can be passed as binaries.\n\n```elixir\n~Z\"\"\"\npub const Packed = packed struct {x: u4, y: u4};\n\npub const Extern = extern struct {x: u16, y: u16};\n\npub fn diff_packed(value: Packed) u8 {\n    return value.x - value.y;\n}\n\npub fn diff_extern(value: Extern) u16 {\n    return value.x - value.y;\n}\n\"\"\"\n\ntest \"packed and extern structs as struct\" do\n  assert 2 = diff_packed(%{x: 7, y: 5})\n  assert 5 = diff_extern(%{x: 47, y: 42})\nend\n\ntest \"packed and extern structs as binary\" do\n  assert 2 = diff_packed(<<0x57>>)\n  assert 5 = diff_extern(<<47::unsigned-size(16)-native, 42::unsigned-size(16)-native>>)\nend\n```\n\n>","ref":"02-collections.html#packed-or-extern-structs"},{"type":"extras","title":"Packed And Extern Endianness {: .warning} - Collection Datatypes","doc":">\n> Be careful about the endianness of packed and extern structs!","ref":"02-collections.html#packed-and-extern-endianness-warning"},{"type":"extras","title":"Pointers to structs. - Collection Datatypes","doc":"Pointers to structs can also be used to marshal data in and out. This is enabled under the\nassumption that you might want to use the struct in a mutable fashion.\n\n```elixir\n~Z\"\"\"\npub fn swap_pointer(value: *Point2D) *Point2D {\n    const temp = value.x;\n    value.x = value.y;\n    value.y = temp;\n\n    return value;\n}\n\"\"\"\n\ntest \"pointer to structs\" do\n  assert %{x: 47, y: 50} == swap_pointer(%{x: 50, y: 47})\nend\n```","ref":"02-collections.html#pointers-to-structs"},{"type":"extras","title":"Nested collections - Collection Datatypes","doc":"The process of marshalling parameters and returns also works with nested array-like and struct data.","ref":"02-collections.html#nested-collections"},{"type":"extras","title":"Arraylike of arraylike - Collection Datatypes","doc":"```elixir\n~Z\"\"\"\npub fn array_of_array_sum(a_of_a: [][]u64) u64 {\n    var sum: u64 = 0;\n    for (a_of_a) |inner_array| {\n        for (inner_array) |value| {\n            sum += value;\n        }\n    }\n    return sum;\n}\n\"\"\"\n\ntest \"array of array\" do\n  assert 21 = array_of_array_sum([[1, 2, 3], [4], [5, 6]])\nend\n```","ref":"02-collections.html#arraylike-of-arraylike"},{"type":"extras","title":"Structs of structs - Collection Datatypes","doc":"```elixir\n~Z\"\"\"\npub const Arrow = struct {\n    head: Point2D,\n    tail: Point2D\n};\n\npub fn reflect_reverse_arrow(arrow: Arrow) Arrow {\n    return .{\n      .head = reflect(arrow.tail), \n      .tail = reflect(arrow.head)\n    };\n}\n\"\"\"\n\ntest \"structs of structs\" do\n  assert %{\n    head: %{x: 1, y: 2},\n    tail: %{x: 3, y: 4}\n  } == reflect_reverse_arrow(%{\n    head: %{x: 4, y: 3},\n    tail: %{x: 2, y: 1}\n  })\nend\n```\n\n> #### Deep Argument Errors {: .info}\n>\n> Argument errors for deeply nested structs will help you understand where your arguments failed to\n> serialize:\n\n```elixir\n# note: skipped because map key order is nondeterministic before 1.15\n@tag [skip: Version.compare(System.version(), \"1.15.0\") == :lt]\ntest \"argument errors\" do\n  assert_raise ArgumentError, \"\"\"\n  errors were found at the given arguments:\n  \n    * 1st argument: \n  \n       expected: map | keyword (for `Arrow`)\n       got: `%{head: %{x: 4, y: 3}, tail: %{x: 2, y: 1.0}}`\n       in field `:tail`:\n       | expected: map | keyword (for `Point2D`)\n       | got: `%{x: 2, y: 1.0}`\n       | in field `:y`:\n       | | expected: integer (for `i32`)\n       | | got: `1.0`\n  \"\"\", fn ->\n    reflect_reverse_arrow(%{\n      head: %{x: 4, y: 3},\n      tail: %{x: 2, y: 1.0}\n    })\n  end\nend\n```","ref":"02-collections.html#structs-of-structs"},{"type":"extras","title":"Arraylike of structs - Collection Datatypes","doc":"```elixir\n~Z\"\"\"\npub fn sum_points(points: []Point2D) Point2D {\n    var result: Point2D = .{.x = 0, .y = 0};\n    for (points) |point| {\n        result.x += point.x;\n        result.y += point.y;\n    }\n    return result;\n}\n\"\"\"\n\ntest \"array of struct\" do\n  assert %{x: 9, y: 12} = sum_points([%{x: 1, y: 2}, %{x: 3, y: 4}, %{x: 5, y: 6}])\nend\n```","ref":"02-collections.html#arraylike-of-structs"},{"type":"extras","title":"Structs of arraylikes - Collection Datatypes","doc":"```elixir\n~Z\"\"\"\npub const PointSOA = struct{\n  x: []u16,\n  y: []u16\n};\n\npub fn sum_point_soa(points: PointSOA) Point2D {\n    var result: Point2D = .{.x = 0, .y = 0};\n    for (points.x) |x| {\n        result.x += x;\n    }\n    for (points.y) |y| {\n        result.y += y;\n    }\n    return result;\n}\n\"\"\"\n\ntest \"struct of array\" do\n  assert %{x: 9, y: 12} = sum_point_soa(%{x: [1, 3, 5], y: [2, 4, 6]})\nend\n```\n\n>","ref":"02-collections.html#structs-of-arraylikes"},{"type":"extras","title":"Interaction with allocators {: .warning} - Collection Datatypes","doc":">\n> if you directly return a datatype that was allocated, it won't be properly cleaned up. However, it\n> can be properly cleaned up by manually deferring its cleanup after calling\n> [`beam.make`](beam.html#make).\n>\n> Cleanup routines in [nif options](4-nif_options.html) will be introduced in a future release, which\n> will enable protection from these sorts of leaks.\n>\n> For more on allocators, see [allocators](3-allocators.html)\n\n```elixir\n~Z\"\"\"\npub fn leaks() !*Point2D {\n    var point = try beam.allocator.create(Point2D);\n\n    point.x = 47;\n    point.y = 50;\n\n    return point;\n}\n\npub fn no_leak() !beam.term {\n    var point = try beam.allocator.create(Point2D);\n    defer beam.allocator.destroy(point);\n\n    point.x = 47;\n    point.y = 50;\n\n    return beam.make(point, .{});\n}\n\"\"\"\n\ntest \"both leaky and non-leaky struct returns work\" do\n  assert %{x: 47, y: 50} == leaks()\n  assert %{x: 47, y: 50} == no_leak()\nend\n```","ref":"02-collections.html#interaction-with-allocators-warning"},{"type":"extras","title":"Allocations","doc":"# Allocations\n\nZig the language has no offically supported allocator, and the standard library datastructures are\nall allocator-agnostic. \n\nZigler ships with three primary allocators, though you can certainly build allocator strategies *on\ntop* of those allocators.","ref":"03-allocators.html"},{"type":"extras","title":"Basic Allocator - Allocations","doc":"The first allocator is `allocator`. This allocator wraps the [nif allocator](https://www.erlang.org/doc/man/erl_nif.html#enif_alloc) \nprovided by the BEAM in the zig allocator interface. You should generally use this allocator over \n`malloc` because it often saves a syscall by using existing preallocated memory pools, because it \nallows the VM to track how much memory your NIF is using, and possibly gives better memory placement \nto avoid cache misses in your execution thread.\n\n```elixir\n~Z\"\"\"\nconst beam = @import(\"beam\");\n\npub fn allocate_raw(count: usize) !beam.term {\n    const slice = try beam.allocator.alloc(u16, count);\n    defer beam.allocator.free(slice);\n\n    for (slice, 0..) |*entry, index| {\n        entry.* = @intCast(index);\n    }\n    return beam.make(slice, .{});\n}\n\"\"\"\n\ntest \"raw allocator\" do\n  assert [0, 1, 2] = allocate_raw(3)\nend\n```\n\n>","ref":"03-allocators.html#basic-allocator"},{"type":"extras","title":"allocator limitations {: .warning} - Allocations","doc":">\n> because the basic allocator directly wraps the beam allocator, according to the documentation:\n>\n> The returned pointer is suitably aligned for any built-in type that fit (sic) in the allocated\n> memory.\n>\n> attempting to allocate memory aligned to a larger size (e.g. page-aligned allocation) will fail\n> using this allocator.","ref":"03-allocators.html#allocator-limitations-warning"},{"type":"extras","title":"Tracking memory. - Allocations","doc":">","ref":"03-allocators.html#tracking-memory"},{"type":"extras","title":"information in hidden globals {: .warning} - Allocations","doc":">\n> Generally storing information in hidden globals is not a good idea. Here it is done to illustrate\n> the memory usage. A better strategy would be to use [resources](5-resources.html)\n\n```elixir\n~Z\"\"\"\nvar global_zigler: []u8 = undefined;\n\npub fn zigler_alloc() !void {\n    global_zigler = try beam.allocator.alloc(u8, 1_000_000);\n}\n\npub fn zigler_free() void {\n    beam.allocator.free(global_zigler);\n}\n\nconst c_stdlib = @cImport(@cInclude(\"stdlib.h\"));\n\nvar global_cstd: [*c]u8 = undefined;\npub fn c_malloc() void {\n    global_cstd = @ptrCast(c_stdlib.malloc(1_000_000));\n}\n\npub fn c_free() void {\n    c_stdlib.free(global_cstd);\n}\n\"\"\"\n\ntest \"zigler memory is tracked\" do\n  Process.sleep(100)\n  start = :erlang.memory[:total]\n  zigler_alloc()\n  assert :erlang.memory[:total] - start >= 1_000_000\n  zigler_free()\nend\n\ntest \"malloc memory is not tracked\" do\n  Process.sleep(100)\n  start = :erlang.memory[:total]\n  c_malloc()\n  assert :erlang.memory[:total] - start <= 1_000_000\n  c_free()\nend\n```","ref":"03-allocators.html#information-in-hidden-globals-warning"},{"type":"extras","title":"Wide Alignment Allocator - Allocations","doc":"Zigler provides a `wide_alignment_allocator` which allows you to allocate memory ranges that have a \nhigher alignment than the maximum alignment for builtin types. \n\n>","ref":"03-allocators.html#wide-alignment-allocator"},{"type":"extras","title":"memory penalty {: .warning } - Allocations","doc":">\n> Note that using this allocator comes with a memory penalty, so use as a general allocator\n> is not recommended.\n\n```elixir\n~Z\"\"\"\npub fn allocate_large_aligned(count: usize) !usize {\n    const page = try beam.wide_alignment_allocator.allocWithOptions(u8, count, 4096, null);\n    defer beam.wide_alignment_allocator.free(page);\n\n    return @intFromPtr(page.ptr);\n}\n\"\"\"\n\ntest \"aligned allocation\" do\n  assert 0 = rem(allocate_large_aligned(3), 4096)\nend\n```","ref":"03-allocators.html#memory-penalty-warning"},{"type":"extras","title":"General Purpose Allocator - Allocations","doc":"Zigler provides a version of the zig standard library's `GeneralPurposeAllocator` which is built on\ntop of the large allocator. Two advantages of using the general purpose allocator include optimized\nmemory layouts for mixed allocation sizes and the ability to track memory leaks.\n\nThe state of the global general purpose allocator is accessible using `beam.allocator_.general_purpose_allocator_instance`\n\nYou may also create a custom general purpose allocator instance using `beam.make_general_purpose_allocator_instance`,\nwhcih is what happens on a per-nif basis if the nif is checking for leaks.\n\n```elixir\n~Z\"\"\"\npub fn leaks() !bool {\n    const memory = try beam.general_purpose_allocator.alloc(u8, 8);\n    defer beam.general_purpose_allocator.free(memory);\n\n    // note that we haven't freed it yet, that happens on deferral,\n    // which lands after the return call.\n\n    return beam.allocator_.general_purpose_allocator_instance.detectLeaks();\n}\n\npub fn noleak() !bool {\n    const memory = try beam.general_purpose_allocator.alloc(u8, 8);\n    beam.general_purpose_allocator.free(memory);\n    return beam.allocator_.general_purpose_allocator_instance.detectLeaks();\n}\n\"\"\"\n\ntest \"leak checks with general purpose allocator\" do\n  require Logger\n  Logger.warning(\"====== the following leak message is expected: =========== START\")\n  Process.sleep(200)\n  assert leaks()\n  Logger.warning(\"=========================================================== END\")\n\n  refute noleak()\nend\n```","ref":"03-allocators.html#general-purpose-allocator"},{"type":"extras","title":"Custom allocators - Allocations","doc":"Because zigler's allocators conform to zig's allocator interface, you can use any composed allocator\nin the standard library or any composable allocator from an imported zig package.\n\n```elixir\n~Z\"\"\"\npub fn with_arena() !beam.term {\n    const std = @import(\"std\");\n\n    var arena = std.heap.ArenaAllocator.init(beam.allocator);\n    defer arena.deinit();\n\n    const allocator = arena.allocator();\n\n    const slice = try allocator.alloc(u16, 4);\n    defer allocator.free(slice);\n\n    for (slice, 0..) |*item, index| {\n        item.* = @intCast(index);\n    }\n\n    return beam.make(slice, .{});\n}\n\"\"\"\n\ntest \"arena allocator\" do\n  assert [0, 1, 2, 3] == with_arena()\nend\n```","ref":"03-allocators.html#custom-allocators"},{"type":"extras","title":"Custom allocators in `beam.get` - Allocations","doc":"If you choose to use a custom allocator, you may use it in the `beam.get` functions to instantiate\ndata where it's the allocator's responsibility to free it at the end.\n\n```elixir\n~Z\"\"\"\npub fn arena_sum(array: beam.term) !u64 {\n    const std = @import(\"std\");\n\n    var arena = std.heap.ArenaAllocator.init(beam.allocator);\n    defer arena.deinit();\n    const arena_allocator = arena.allocator();\n\n    const slice = try beam.get([]u64, array, .{.allocator = arena_allocator});\n\n    var total: u64 = 0;\n\n    for (slice) |item| { total += item; }\n\n    return total;\n}\n\"\"\"\n\ntest \"arena sum\" do\n  assert 6 == arena_sum([1, 2, 3])\nend\n```","ref":"03-allocators.html#custom-allocators-in-beam-get"},{"type":"extras","title":"Nif options","doc":"# Nif options\n\nZigler gives you several ways to configure compilation options for nifs. These options are part of\nthe `use Zig` directive, in the keyword options under the key `:nifs`. This key itself points to a\nkeyword list where the keys are the names of the nif functions and the values are a keyword list of\noptions that apply to that function.\n\nThis guide shows you all of the nif options except for options related to\n[Resources](5-resources.html), [C integration](6-c_integration.html), or\n[Concurrency](7-concurrency.html).","ref":"04-nif_options.html"},{"type":"extras","title":"Automatic options (elixir) - Nif options","doc":"To declare that functions should have their options automatically determined, use `...` in the nifs\nparameter list. Nifs which sholud have manually decided options should come after the `...` as a\nkeyword list. If all options are automatically determined, then omitting the `:nif` keyword\ncompletely is valid.\n\n```elixir\ndefmodule AutomaticOptions do\n  use Zig, \n    otp_app: :zigler,\n    nifs: [...]\n\n  ~Z\"\"\"\n  pub fn noop() void {}\n  \"\"\"\nend\n```","ref":"04-nif_options.html#automatic-options-elixir"},{"type":"extras","title":"Automatic options (erlang) - Nif options","doc":"Erlang cannot interpret the `...` AST in elixir, so you must use `[auto]` atom in the nifs options\ninstead.\n\n```erlang\n-zig_opts([{nifs, [auto]}])\n```","ref":"04-nif_options.html#automatic-options-erlang"},{"type":"extras","title":"Return type - Nif options","doc":"In [Collections](2-collections.html) we saw how certain collection types could be manually\nmarshalled into alternative representations using [`beam.make`](beam.html#make). This can be handled\nas a nif configuration as follows. The advantage to doing it this way is that the typespec for the\nfunction will correctly reflect the return type.\n\n```elixir\ndefmodule ReturnTypeTest do\n  use ExUnit.Case, async: true\n\n  use Zig, \n    otp_app: :zigler,\n    nifs: [\n        returns_binary: [return: :binary],\n        returns_list: [return: :list]\n    ]\n\n  ~Z\"\"\"\n  pub fn returns_binary() [3]u16 {\n      return [3]u16{47, 48, 49};\n  }\n\n  pub fn returns_list() []const u8 {\n      return \"Hello world!\";\n  }\n  \"\"\"\n\n  test \"returns binary\" do\n    assert <<47, 0, 48, 0, 49, 0>> = returns_binary()\n  end\n\n  test \"returns list\" do\n    assert ~C'Hello world!' = returns_list()\n  end\nend\n```","ref":"04-nif_options.html#return-type"},{"type":"extras","title":"Alias - Nif options","doc":"It's possible to create a new function which is an alias of another function. This is how it's done:\n\n```elixir\ndefmodule AliasTest do\n  use ExUnit.Case, async: true\n\n  use Zig, \n    otp_app: :zigler,\n    nifs: [\n      ...,\n      new_function: [alias: :old_function]\n    ]\n\n  ~Z\"\"\"\n  pub fn old_function() u32 {\n      return 47;\n  }\n  \"\"\"\n\n  test \"both main and alias functions work\" do\n    assert 47 == old_function()\n    assert 47 == new_function()\n  end\nend\n```","ref":"04-nif_options.html#alias"},{"type":"extras","title":"Args options - Nif options","doc":"Arguments can also take options, using `args: [...]`","ref":"04-nif_options.html#args-options"},{"type":"extras","title":"Noclean - Nif options","doc":"If you want to disable automatic allocator cleanup of datatypes, you can do so either in the\n`:return` or `:args` section by including the `:noclean` option. This is most useful if your args or\nreturn data are going to be persisted beyond the lifetime of the nif call, though doing this in many\ncases is not recommended.\n\nThis flag can be stacked with previous options, for example: `return: [:noclean, :binary]`.","ref":"04-nif_options.html#noclean"},{"type":"extras","title":"Leak Check - Nif options","doc":"It's possible to wrap each function call in its own instance of\n[`beam.general_purpose_allocator`](beam.html#general_purpose_allocator) bound into the\n[`beam.allocator`](beam.html#allocator) threadlocal variable. If you tag your nif as `leak_check`,\nit will check that `beam.allocator` has cleared all of its contents at the end of the function call,\nand if that hasn't happened, it raises.\n\n```elixir\ndefmodule LeakCheckTest do\n  use ExUnit.Case, async: true\n\n  use Zig,\n    otp_app: :zigler,\n    nifs: [check_me: [leak_check: true]]\n\n  ~Z\"\"\"\n  const beam = @import(\"beam\");\n  pub fn check_me() !void {\n      _ = try beam.context.allocator.create(u8);\n  }\n  \"\"\"\n\n  test \"leak check\" do\n    require Logger\n    Logger.warning(\"====== the following leak message is expected: =========== START\")\n    Process.sleep(200)\n    assert_raise RuntimeError, \"memory leak detected in function `LeakCheckTest.check_me/0`\", fn ->\n      check_me()\n    end\n    Logger.warning(\"=========================================================== END\")\n  end\nend\n```\n\n`leak_check` can also be applied to all nifs in the module:\n\n```elixir\ndefmodule LeakCheckAllTest do\n  use ExUnit.Case, async: true\n\n  use Zig,\n    otp_app: :zigler,\n    leak_check: true\n\n  ~Z\"\"\"\n  const beam = @import(\"beam\");\n  pub fn check_me() !void {\n      _ = try beam.context.allocator.create(u8);\n  }\n  \"\"\"\n\n  test \"leak check\" do\n    require Logger\n    Logger.warning(\"====== the following leak message is expected: =========== START\")\n    Process.sleep(200)\n\n    assert_raise RuntimeError, \"memory leak detected in function `LeakCheckAllTest.check_me/0`\", fn ->\n      check_me()\n    end\n    Logger.warning(\"=========================================================== END\")\n  end\nend\n```","ref":"04-nif_options.html#leak-check"},{"type":"extras","title":"Typespec override - Nif options","doc":"Typespecs generation can be suppressed by setting `spec: false`.\nIf you want typespecs for such functions, specify using `@spec` elsewhere\nin your module.\n\nFor example:\n\n```elixir\ndefmodule Override do\n  use Zig, \n    otp_app: :zigler,\n    nifs: [typespec_override: [spec: false]]\n\n  @spec typespec_override(integer) :: integer\n  ~Z\"\"\"\n  const beam = @import(\"beam\");\n  pub fn typespec_override(term: beam.term) !beam.term {\n      const input = try beam.get(u32, term, .{});\n      return beam.make(input + 1, .{});\n  }\n  \"\"\"\nend\n```","ref":"04-nif_options.html#typespec-override"},{"type":"extras","title":"Disable documentation - Nif options","doc":"Documentation can be disabled with the `docs: false` option.\n\n```elixir\ndefmodule DisableDoc do\n  use Zig, \n    otp_app: :zigler,\n    nifs: [nodocs: [docs: false]]\n\n  ~Z\"\"\"\n  pub fn nodocs() void {}\n  \"\"\"\nend\n#module\n```","ref":"04-nif_options.html#disable-documentation"},{"type":"extras","title":"Resources","doc":"# Resources\n\nResources are datatypes which are managed by the BEAM reference-counted garbage collector. If you\nare passing data between function calls, generally it is best practice to pass them as a resource\ninstead of as a pointer or a global variable.\n\nFor documentation of how resources work in general, see the erlang\n[documentation](https://www.erlang.org/doc/man/erl_nif.html#resource_objects) on the topic. Note\nthat the conveniences in [beam.Resource](beam.html#Resource) generic type exist to make operations\ntype-safe.\n\n>","ref":"05-resources.html"},{"type":"extras","title":"Resources are references {: .info} - Resources","doc":">\n> In managed language environments resources are passed as `t:reference/0` tokens.\n\nUnder the hood the BEAM allocates a chunk of memory on resource creation, and this memory is cleared\nwhen the garbage collector is triggered. The BEAM also allows you to add in trigger hooks for when\nthis event happens, so custom cleanup can be tied into these GC events.\n\n>","ref":"05-resources.html#resources-are-references-info"},{"type":"extras","title":"Passing references between modules {: .warning} - Resources","doc":">\n> This is not currently supported by Zigler, but support is planned.","ref":"05-resources.html#passing-references-between-modules-warning"},{"type":"extras","title":"Declaring a resource - Resources","doc":"In order to use a resource, you must do at a minimum three things:\n\n1. declare the resource in your `use Zig` directive, under the `:resource` key. This is a list of\n  atoms, which match the name of the resource type.\n1. declare the wrapped type. This is the type of the data that is placed in the memory space of the\n  resource.\n1. use [`beam.Resource`](beam.html#Resource) to declare the resource type. note that this resource type\n  must be `pub`.\n\n```elixir\ndefmodule ResourceTest do\n  use ExUnit.Case, async: true\n  use Zig, \n    otp_app: :zigler,\n    resources: [\n      :StructResource,\n      :PointerResource\n    ]\n\n  ~Z\"\"\"\n  const beam = @import(\"beam\");\n  const root = @import(\"root\");\n\n  const MyStruct = struct {\n      payload: u64\n  };\n\n  pub const StructResource = beam.Resource(MyStruct, root, .{});\n  \"\"\"\n```","ref":"05-resources.html#declaring-a-resource"},{"type":"extras","title":"Using resources in functions - Resources","doc":"```elixir\n~Z\"\"\"\npub fn create_resource_term(number: u64) !beam.term {\n    const res = try StructResource.create(.{.payload = number}, .{});\n    return beam.make(res, .{});\n}\n\npub fn retrieve_resource_term(term: beam.term) !u64 {\n    const res = try beam.get(StructResource, term, .{});\n    return res.unpack().payload;\n}\n\"\"\"\n\ntest \"lifecyle operations through terms\" do\n  resource = create_resource_term(47)\n  assert is_reference(resource)\n  assert 47 = retrieve_resource_term(resource)\nend\n```\n\nResources can be marshalled into and out of [`beam.term`](beam.html#term) values using\n[`beam.make`](beam.html#make) and [`beam.get`](beam.html#get) functions as with any other types.\n\nIn order to convert between the resource type and the wrapped type, you'll need to use the `create`\nand `unpack` functions. Note that `create` is failable since under the hood it uses the BEAM\nresource allocator.","ref":"05-resources.html#using-resources-in-functions"},{"type":"extras","title":"Direct marshalling - Resources","doc":"It's possible to directly return resources from a nif function and also pass them as parameters: The\nnif marshalling functions will be able to detect these types and assign them correctly.\n\n```elixir\n~Z\"\"\"\npub fn create_resource_direct(number: u64) !StructResource {\n    return StructResource.create(.{.payload = number}, .{});\n}\n\npub fn retrieve_resource_direct(resource: StructResource) u64 {\n    return resource.unpack().payload;\n}\n\"\"\"\n\ntest \"direct lifecyle operations\" do\n  resource = create_resource_direct(47)\n  assert is_reference(resource)\n  assert 47 = retrieve_resource_direct(resource)\nend\n\ntest \"must be the correct type of reference\" do\n  assert_raise ArgumentError, \"\"\"\n  errors were found at the given arguments:\n\n    * 1st argument: \n  \n       expected: reference (for `beam.Resource(MyStruct, @import(\\\"root\\\"), .{...})`)\n       got: `%{payload: 42}`\n  \"\"\", fn ->\n    retrieve_resource_direct(%{payload: 42})\n  end\n\n  non_resource_ref = make_ref()\n\n  message = \"\"\"\n  errors were found at the given arguments:\n  \n    * 1st argument: \n  \n       expected: reference (for `beam.Resource(MyStruct, @import(\\\"root\\\"), .{...})`)\n       got: `#{inspect non_resource_ref}`\n       note: the reference passed is not associated with a resource of the correct type\n  \"\"\"\n\n  assert_raise ArgumentError, message, fn ->\n    retrieve_resource_direct(non_resource_ref)\n  end\nend\n```\n\n>","ref":"05-resources.html#direct-marshalling"},{"type":"extras","title":"no coercion {: .warning} - Resources","doc":">\n> You can't pass a term of the same type as the wrapped type and use it within the function.","ref":"05-resources.html#no-coercion-warning"},{"type":"extras","title":"Wrapping pointers and cleanup - Resources","doc":"In many cases you won't want to move large data structures into or out of the resource-allocated\nmemory space; this incurs a data copy cost. In that case, you might want to store a pointer in the\nresource memory space.\n\nIn order to do properly clean up after this, you'll need to write a callback function and store it\nin a struct namespace that gets associated with the resource type in using the `beam.Resource`\nfunction `.Callbacks` option.\n\nNote that the destructor can also be used in cases where other resources need to be cleaned up, for\nexample file descriptors.\n\nThe following functions are supported in the Callbacks, and are all optional.\n\n- `dtor`: called when the GC collects the \n- `stop`: called on stop on behalf of `e.enif_select`\n- `down`: called on resource down, on behalf of `e.enif_monitor_process`\n- `dyncall`: called on dynamic resource call, on behalf of `enif_dynamic_resource_call`\n\n```elixir\n~Z\"\"\"\npub const PointerResource = beam.Resource(*MyStruct, root, .{.Callbacks = PointerResourceCallbacks});\n\npub const PointerResourceCallbacks = struct {\n    pub fn dtor(s: **MyStruct) void {\n        beam.allocator.destroy(s.*);\n    }\n};\n\npub fn create_pointer_resource(number: u64) !PointerResource {\n    const new_struct = try beam.allocator.create(MyStruct);\n    new_struct.payload = number;\n    return PointerResource.create(new_struct, .{});\n}\n\npub fn retrieve_pointer_resource(resource: PointerResource) u64 {\n    return resource.unpack().*.payload;\n}\n\"\"\"\n\ntest \"pointer-based lifecyle operations\" do\n  resource = create_pointer_resource(47)\n  assert is_reference(resource)\n  assert 47 = retrieve_pointer_resource(resource)\nend\n```\n\n>","ref":"05-resources.html#wrapping-pointers-and-cleanup"},{"type":"extras","title":"pointer allocation strategy {: .warning} - Resources","doc":">\n> It is strongly recommended to use [`beam.allocator`](beam.html#allocator) for your pointer payload\n> allocators, as [`beam.allocator`](beam.html#allocator) is undefined in the callback context.\n>\n> alternatively, if you do use your own managed allocator, you can pack a pointer to the allocator\n> into your datastructure and use this pointer to clean up.","ref":"05-resources.html#pointer-allocation-strategy-warning"},{"type":"extras","title":"Release and Keep - Resources","doc":"Resources are associated with `release` and `keep` functionality. These increment and decrement the\nreference count on the resource which allows a nif to prevent the GC from destroying the memory and\ncalling the destructor. Normally the nif management functions select default `release` and `keep`\nsettings so that the functions do the most expected outcome - nif functions keep resources while\nrunning and then release them when they're finished.\n\n>","ref":"05-resources.html#release-and-keep"},{"type":"extras","title":"release on creation {: .info} - Resources","doc":">\n> A Struct resource is normally released on creation. This can be disabled by passing `.release =\n> false` into the options parameter of `beam.Resource(...).create(...)`\n\n>","ref":"05-resources.html#release-on-creation-info"},{"type":"extras","title":"keep on get {: .info} - Resources","doc":">\n> By default, [`beam.get`](beam.html#get) will keep the resource when the internal term retrieved.\n> This can be disabled by passing `.keep = false` into the options parameter of `beam.get`\n\n>","ref":"05-resources.html#keep-on-get-info"},{"type":"extras","title":"get without keeping {: .warning} - Resources","doc":">\n> for wrapped datatypes that require a cleanup step (e.g. pointers) it is not recommended to get\n> without keeping, as there could be a race condition where dereferencing the pointer occurs after\n> another nif running in a different OS thread has performed a cleanup against the same resource.\n\n>","ref":"05-resources.html#get-without-keeping-warning"},{"type":"extras","title":"release when function argument {: .info} - Resources","doc":">\n> If a function is passed a `beam.Resource(...)` type, it will release it at the end of the call. This\n> can be disabled by setting `:noclean` flag in the function argument options. (see [nif\n> options](4-nif_options.html#noclean))\n\nThese functions are provided in the resource type as `beam.Resource(...).release(...)` and\n`beam.Resource(...).keep(...)` functions, respectively:\n\n```elixir\n~Z\"\"\"\npub fn release(resource: StructResource) void {\n    resource.release();\n}\n\npub fn keep(resource: StructResource) void {\n    resource.keep();\n}\n\"\"\"\n```","ref":"05-resources.html#release-when-function-argument-info"},{"type":"extras","title":"CXX integration","doc":"# CXX integration\n\nZigler offers several tools to integrate your code with C and C++ code.","ref":"06-c_integration.html"},{"type":"extras","title":"compiling C using the C toolchain - CXX integration","doc":"If you want to compile C or C++ files using the C and C++ toolchain bundled with the zig programming\nlanguage, you should include a path to your include directory and a source path or list of source\npaths to be compiled, via the `include_dir` and `src` module options.\n\n>","ref":"06-c_integration.html#compiling-c-using-the-c-toolchain"},{"type":"extras","title":"source paths {: .info} - CXX integration","doc":">\n> source paths may contain a trailing *, which will pull all c or c++ files in that directory, (but\n> not subdirectory)","ref":"06-c_integration.html#source-paths-info"},{"type":"extras","title":"included.h - CXX integration","doc":"```c\n// forwarded function definition\nint plus_one(int);\n```","ref":"06-c_integration.html#included-h"},{"type":"extras","title":"src/src.c - CXX integration","doc":"```c\nint plus_one(int value) {\n    return value + 1;\n}\n```","ref":"06-c_integration.html#src-src-c"},{"type":"extras","title":"Elixir code - CXX integration","doc":"```elixir\ndefmodule CompilingC do\n  use ExUnit.Case, async: true\n  use Zig, \n    otp_app: :zigler,\n    c: [include_dirs: \"include\", src: \"src/*\"] \n\n  ~Z\"\"\"\n  const c = @cImport(@cInclude(\"included.h\"));\n\n  pub const plus_one = c.plus_one;\n  \"\"\"\n\n  test \"c plus one\" do\n    assert 48 = plus_one(47)\n  end\nend\n```","ref":"06-c_integration.html#elixir-code"},{"type":"extras","title":"linking against a C abi library - CXX integration","doc":"This example shows you how to link in a system library (which can be `.a`, `.so`, `.obj`, or\n`.dll`). Zig will resolve the extension based on the operating system native rules. To use the\nfunctions in the library, there must also be an associated `.h` file with `extern` functions.\n\nIn this example we'll use the `cblas_dasum` function, which takes a length, an pointer to\ndouble-precision floating point list, and a integer stride. The result is a sum of the numbers in\nthe list.\n\nThe [rules for collections](#2-collections.html) apply to functions that are directly imported from\nC files.\n\n```elixir\nif {:unix, :linux} == :os.type() do\n# currently we only have access to the BLAS library on linux CI actions, so\n# it's unavailable for other operating systems for automated testing purposes\n\n  defmodule LibraryTest do\n    use ExUnit.Case, async: true\n    use Zig, \n      otp_app: :zigler,\n      c: [link_lib: {:system, \"blas\"}]\n\n    ~Z\"\"\"\n    pub const dasum = @cImport(@cInclude(\"cblas.h\")).cblas_dasum;\n    \"\"\"\n\n    test \"dasum\" do\n      assert 6.0 == dasum(3, [1.0, 2.0, 3.0], 1)\n    end\n  end\n```\n\n>","ref":"06-c_integration.html#linking-against-a-c-abi-library"},{"type":"extras","title":"linking against libcpp: {: .info} - CXX integration","doc":">\n> if you need to link against `libcpp`, the library has a special-cased option: `link_libcpp: true`","ref":"06-c_integration.html#linking-against-libcpp-info"},{"type":"extras","title":"Easy-C - CXX integration","doc":"It's also possible to also automatically create nifs without writing zig function shims. This works\neither with linking an external library with `link_lib` or building your own code with `src`.\n\nBecause the C ABI exposes all functions publically in a global namespace, we can't use zigler's\nautomatic detection to decide which functions to surface.\n\nHere are the steps to using `easy_c`:\n\n- declare which header file you'd like to use with the `easy_c` module option\n  - you may want to add `include_dir` if the header isn't a system C header.\n- add a `link_lib` or `src` option to make sure that the functions are built.\n- declare which functions you'd like to hoist into the module.\n\nIn this example we'll use the `cblas_daxpy` function, which takes a length, a double-precision `a`\nvalue, a list (`x`) of double-precision values, a stride for the `x` values, a list (`y`) of\ndouble-precision values, and a stride for `y`. \n\nIt then calculates the linear transformation `ax + y` using these vectors.\n\n> ### \"in-out\" parameters {: .info}\n>\n> The `y` parameter in this function is an \"in-out\" parameter. Instead of returning a result, the\n> results are written into the y pointer.\n>\n> to specify that it's an in-out parameter, we put an integer (which is the 0-indexed index of the\n> argument which is the out parameter).\n>\n> since the returned pointer doesn't have a specified value, we have to specify a `length` option,\n> which could either be a fixed number, or `{:arg, n}` where `n` is the argument index corresponding\n> to a variable length.\n>\n> We can also use [nif options like alias](4-nif_options.html#alias) and [the return option\n> `binary`](2-collections.html#selecting-output-type) alongside these special `easy_c` options.\n\n```elixir\n  defmodule EasyCTest do\n    use ExUnit.Case, async: true\n    use Zig, \n      otp_app: :zigler,\n      easy_c: \"cblas.h\",\n      c: [link_lib: {:system, \"blas\"}],\n      nifs: [\n        cblas_daxpy: [params: %{4 => :in_out}, return: [length: {:arg, 0}]],\n        cblas_daxpy_bin: [alias: :cblas_daxpy, params: %{4 => :in_out}, return: [:binary, length: {:arg, 0}]]\n      ]\n\n    test \"daxpy as a list\" do\n      assert [7.0, 11.0, 15.0] == cblas_daxpy(3, 3.0, [1.0, 2.0, 3.0], 1, [4.0, 5.0, 6.0], 1)\n    end\n\n    test \"daxpy as a binary\" do\n      assert <<7.0::float-native, 11.0::float-native, 15.0::float-native>> ==\n               cblas_daxpy_bin(3, 3.0, [1.0, 2.0, 3.0], 1, [4.0, 5.0, 6.0], 1)\n    end\n  end\nend\n# module\n```","ref":"06-c_integration.html#easy-c"},{"type":"extras","title":"Nif concurrency strategies","doc":"# Nif concurrency strategies\n\nWhen execution flow enters a Nif, control is fully relinquished from the managed environment of the\nBEAM VM to a context where the BEAM is more or less unaware of what is going on.\n\nIn general the VM cannot tolerate native code running for [longer than approximately one\nmillisecond](https://www.erlang.org/doc/man/erl_nif.html#lengthy_work).\n\nThere are several tools that the BEAM nif system provides for you to","ref":"07-concurrency.html"},{"type":"extras","title":"Synchronous - Nif concurrency strategies","doc":"The default mode for Nifs to run is *synchronous*. Only use this mode if you are confident that your\ncode can run in under 1ms.","ref":"07-concurrency.html#synchronous"},{"type":"extras","title":"Dirty CPU - Nif concurrency strategies","doc":"`dirty_cpu` mode is usable when your VM has created *Dirty CPU schedulers*. By default, the VM\ncreates one dirty CPU scheduler per CPU core available to it. Nifs tagged as `dirty_cpu` are allowed\nto run longer than 1 millisecond.\n\nIn order to tag a function as `dirty_cpu`, use the `:dirty_cpu` flag in the options list for the\nfunction in the `:nif` call.","ref":"07-concurrency.html#dirty-cpu"},{"type":"extras","title":"beam.yield in Dirty CPU - Nif concurrency strategies","doc":"The [`beam.yield`](beam.html#yield) function in dirty CPU mode will detect if the parent process has\ndied and will return `error.processterminated`.\n\n```elixir\ndefmodule DirtyCpu do\n  use ExUnit.Case, async: true\n  use Zig, \n    otp_app: :zigler,\n    nifs: [long_running: [:dirty_cpu]]\n\n  ~Z\"\"\"\n  const beam = @import(\"beam\");\n  // this is a dirty_cpu nif.\n  pub fn long_running(pid: beam.pid) !void {\n      defer {\n        // code in the defer block is triggered when process is killed.\n        // we need to create a new thread-independent context because\n        // the context from the running process is now invalid.\n        const env = beam.alloc_env();\n        beam.send(pid, .killed, .{.env = env}) catch unreachable;\n        beam.free_env(env);\n      }\n\n      try beam.send(pid, .unblock, .{});\n\n      while(true) {\n          try beam.yield();\n      }\n  }\n  \"\"\"\n\n  test \"dirty cpu can be cancelled\" do\n    this = self()\n    dirty_cpu = spawn(fn -> long_running(this) end)\n    assert_receive :unblock\n    Process.exit(dirty_cpu, :kill)\n    assert_receive :killed\n  end\nend\n```\n\n>","ref":"07-concurrency.html#beam-yield-in-dirty-cpu"},{"type":"extras","title":"queue limitations {: .warning} - Nif concurrency strategies","doc":">\n> if you consume all of your dirty cpu schedulers with nif calls, the next `dirty_cpu` call will block\n> until a scheduler frees up; this could cause undesired latency characteristics.","ref":"07-concurrency.html#queue-limitations-warning"},{"type":"extras","title":"Dirty IO - Nif concurrency strategies","doc":"It's not recommended to use `dirty_io` unless you're performing IO operations and blocking using nif\nevents and blocking operations. \n\nIn order to tag a function as `dirty_io`, use the `:dirty_io` flag in the options list for the\nfunction in the `:nif` call.","ref":"07-concurrency.html#dirty-io"},{"type":"extras","title":"Threaded - Nif concurrency strategies","doc":"`threaded` mode is usable when your OS supports spawning threads. This is effectively all current\nplatforms supporting the BEAM VM today. Zigler will wrap your function code \n\nIn order to tag a function as `threaded`, use the `:threaded` flag in the options list for the\nfunction in the `:nif` call. Generally, no other changes must be made to execute a function in\nthreaded mode.\n\n>","ref":"07-concurrency.html#threaded"},{"type":"extras","title":"env in Threaded mode {: .warning} - Nif concurrency strategies","doc":">\n> The `env` variable when you run in threaded mode is *not* a process-bound environment.","ref":"07-concurrency.html#env-in-threaded-mode-warning"},{"type":"extras","title":"beam.yield in Threaded mode - Nif concurrency strategies","doc":"The [`beam.yield`](beam.html#yield) function in dirty CPU mode will detect if the parent process has\ndied and will return `error.processterminated`.\n\n>","ref":"07-concurrency.html#beam-yield-in-threaded-mode"},{"type":"extras","title":"return from yield quickly! {: .warning} - Nif concurrency strategies","doc":">\n> You must return from the yield quickly (within 750us). If you are unable to return quickly, then\n> zigler run will cause the thread metadata to leak. This will be fixed in zigler 0.11.\n\n```elixir\ndefmodule Threaded do\n  use ExUnit.Case, async: true\n  use Zig, \n    otp_app: :zigler,\n    nifs: [long_running: [:threaded]]\n\n  ~Z\"\"\"\n  const beam = @import(\"beam\");\n  const std = @import(\"std\");\n  pub fn long_running(pid: beam.pid) !void {\n      // following code triggered when process is killed.\n\n      defer {\n        beam.send(pid, .killed, .{}) catch {};\n      }\n\n      while(true) {\n          _ = try beam.send(pid, .unblock, .{});\n          try beam.yield();\n      }\n  }\n  \"\"\"\n\n  @tag :threaded\n  test \"threaded can be cancelled\" do\n    this = self()\n    threaded = spawn(fn -> long_running(this) end)\n    #assert_receive :unblock\n    Process.sleep(100)\n    Process.exit(threaded, :kill)\n    assert_receive :killed\n    Process.sleep(1000)\n  end\nend\n```","ref":"07-concurrency.html#return-from-yield-quickly-warning"},{"type":"extras","title":"Yielding - Nif concurrency strategies","doc":">","ref":"07-concurrency.html#yielding"},{"type":"extras","title":"yielding nifs {: .error} - Nif concurrency strategies","doc":">\n> Yielding nifs are not available in this release of Zigler\n\n```elixir\n# module\n```","ref":"07-concurrency.html#yielding-nifs-error"},{"type":"extras","title":"Global module options","doc":"# Global module options","ref":"08-module_options.html"},{"type":"extras","title":"ignore - Global module options","doc":"Public functions can be ignored and *not* converted into nifs by filling out the `:ignore` option in\n`use Zig` directive.\n\n```elixir\ndefmodule IgnoreTest do\n  use ExUnit.Case, async: true\n\n  use Zig, \n    otp_app: :zigler,\n    ignore: [:ignored]\n\n  ~Z\"\"\"\n  pub fn ignored(number: u32) u32 {\n      return number + 1;\n  }\n\n  pub fn available(number: u32) u32 {\n      return ignored(number);\n  }\n  \"\"\"\n\n  test \"available function works\" do\n    assert 48 = available(47)\n  end\n\n  test \"ignored function is not available\" do\n    refute function_exported?(__MODULE__, :ignored, 0)\n  end\nend\n```","ref":"08-module_options.html#ignore"},{"type":"extras","title":"attributes - Global module options","doc":"Attributes from your module can be used as compile-time constants communicated \nfrom elixir.  All attributes of the following types will be automatically \navailable through the `attributes` module import:\n\n- `integer` (as `comptime int` values)\n- `float` (as `comptime float` values)\n- `nil` (as `null`)\n- `boolean` (as `bool` values)\n- `binary` (as `comptime [:0]u8` values)\n- `atom` (as *enum literal* values)\n- `tuple` of only the above types (as *tuple*)\n\n```elixir\ndefmodule Attribute do\n  use ExUnit.Case, async: true\n  use Zig, otp_app: :zigler\n\n  @supplied_value Mix.env()\n\n  ~Z\"\"\"\n  const beam = @import(\"beam\");\n  const attribs = @import(\"attributes\");\n\n  pub fn get_attrib() beam.term {\n    return beam.make(.{.ok, attribs.supplied_value}, .{});\n  }\n  \"\"\"\n\n  test \"getting an attribute\" do\n    assert {:ok, :test} = get_attrib()\n  end\nend\n```","ref":"08-module_options.html#attributes"},{"type":"extras","title":"adding packages - Global module options","doc":"It's possible to add zig files as packages using the `packages` keyword option. The name of the\npackage is the key, and the value is a tuple of the path to the zig file that acts as the package\nand a list of dependencies for the package.","ref":"08-module_options.html#adding-packages"},{"type":"extras","title":"Example extra.zig - Global module options","doc":"```zig\npub const value = 47;\n```\n\n```elixir\ndefmodule PackageFile do\n  use ExUnit.Case, async: true\n  use Zig, \n    otp_app: :zigler,\n    packages: [extra: {\"test/_support/package/extra.zig\", [:beam]}]\n\n  ~Z\"\"\"\n  const extra = @import(\"extra\");\n\n  pub fn extra_value() u64 {\n    return extra.value;\n  }\n  \"\"\"\n\n  test \"package file\" do\n    assert 47 = extra_value()\n  end\nend\n#module\n```","ref":"08-module_options.html#example-extra-zig"},{"type":"extras","title":"dump options - Global module options","doc":"Zigler lets you dump various compile-time assets to the console for\ndebugging purposes, which can be enabled by setting any given one\nof the following options to `true`:\n\n- `dump`: dumps the rendered elixir code generated by `use Zig`.\n- `dump_sema`: dumps the json data emitted by the semantic analysis pass.\n- `dump_build_zig`: dumps the autogenerated `build.zig` file","ref":"08-module_options.html#dump-options"},{"type":"extras","title":"Raw nifs","doc":"# Raw nifs\n\nAll of the nifs shown to this point involve zigler constructing an adapter function with\nterm marshalling automatically generated.  It is also possible to run a nif without\ndoing any of those steps.\n\nThe normal (C) header for a BEAM nif is as follows:\n\n`static ERL_NIF_TERM hello(ErlNifEnv* env, int argc, const ERL_NIF_TERM argv[])`\n\nZigler can interpret certain headers as intended for being called as a raw nif.","ref":"09-raw_nifs.html"},{"type":"extras","title":"Allowed Signatures - Raw nifs","doc":"The following two zig function signatures are interpreted as raw nifs:\n\n- `fn (beam.env, c_int, [*]beam.term) beam.term`\n- `fn (beam.env, c_int, [*]e.ErlNifTerm) e.ErlNifTerm`","ref":"09-raw_nifs.html#allowed-signatures"},{"type":"extras","title":"nif options setup - Raw nifs","doc":"A raw nif MUST contain the `arity` option.  This can be one of:\n\n- a single integer, representing the desired arity of the nif function\n- a single range, representing a range of arities for the nif function\n- a list of integers and rannges, representing all the arities of the nif function.","ref":"09-raw_nifs.html#nif-options-setup"},{"type":"extras","title":"Example - Raw nifs","doc":"```elixir\ndefmodule RawCallTest do\n  use ExUnit.Case, async: true\n\n  use Zig, \n    otp_app: :zigler,\n    nifs: [\n        raw_call_beam: [arity: 1],\n        raw_call_erl_nif: [arity: 1],\n        raw_call_multi_arity: [arity: [0, 2..3]]\n    ]\n\n  ~Z\"\"\"\n  const beam = @import(\"beam\");\n  const e = @import(\"erl_nif\");\n\n  pub fn raw_call_beam(env: beam.env, count: c_int, list: [*]const beam.term) beam.term {\n      return beam.make(.{.count = count, .item = list[0]}, .{.env = env});\n  }\n\n  pub fn raw_call_erl_nif(env: beam.env, count: c_int, list: [*]const e.ErlNifTerm) e.ErlNifTerm {\n      return beam.make(.{.count = count, .item = beam.term{.v = list[0]}}, .{.env = env}).v;\n  }\n\n  pub fn raw_call_multi_arity(env: beam.env, arity: c_int, _: [*]const beam.term) beam.term {\n      return beam.make(arity, .{.env = env});\n  }\n  \"\"\"\n\n  test \"raw call with beam format\" do\n    assert %{count: 1, item: {:foo, \"bar\"}} = raw_call_beam({:foo, \"bar\"})\n  end\n  \n  test \"raw call with erl_nif format\" do\n    assert %{count: 1, item: {:foo, \"bar\"}} = raw_call_erl_nif({:foo, \"bar\"})\n  end\n\n  test \"raw call with multiple arities\" do\n    assert 0 = raw_call_multi_arity()\n    assert 2 = raw_call_multi_arity(:foo, :bar)\n    assert 3 = raw_call_multi_arity(:foo, :bar, :baz)\n\n    refute function_exported?(__MODULE__, :raw_call_multi_arity, 1)\n  end\nend\n#module\n```\n\n>","ref":"09-raw_nifs.html#example"},{"type":"extras","title":"beam.make and beam.get in raw nifs {: .warning} - Raw nifs","doc":">\n> Note that you MUST supply `.{.env = env}` in the options to beam.make or beam.get calls in raw nifs,\n> or functions called by raw nifs. The threadlocal `beam.context` variable which normally stores the\n> environment is not set when you make a raw call.","ref":"09-raw_nifs.html#beam-make-and-beam-get-in-raw-nifs-warning"},{"type":"extras","title":"Module Callbacks","doc":"# Module Callbacks\n\nThe BEAM provides for a module to have several event-callbacks that are fired \nwhen a module is created.  Zigler labels these callbacks in the following way:\n\n- [`on_load`](#on_load) - when the module is being loaded\n- [`on_upgrade`](#on_upgrade) - when the module is being loaded to replace another module\n- [`on_unload`](#on_unload) - when the module is being purged\n\nThese labeled are passed as options to the `use Zig` directive with the\nname of the function to be used as the callback.  All three are optional.\nYou may use just the atom to stand a function with the same name, e.g:\n`callbacks: [:on_load]` is shorthand for `callbacks: [on_load: :on_load]`\n\n>","ref":"10-callbacks.html"},{"type":"extras","title":"Callbacks are pub {: .info } - Module Callbacks","doc":">\n> Be sure to make your callback functions `pub`.  You do *not* need to\n> add them to the module's `ignore` option.\n\n>","ref":"10-callbacks.html#callbacks-are-pub-info"},{"type":"extras","title":"Context in callbacks {: .info } - Module Callbacks","doc":"> \n> For all callbacks, the context is set as follows:\n> \n> - `env`: the `e.ErlNifEnv` value passed to the callback function;\n> - `mode`: `.callback`;\n> - `allocator`: `beam.allocator`;\n> \n> Thus you should be able to use `beam.get`, `beam.make`, or `beam.send` \n> with the appropriate context set without extra options.","ref":"10-callbacks.html#context-in-callbacks-info"},{"type":"extras","title":"on_load - Module Callbacks","doc":"The on_load callback may have one of the following function signatures:\n\n- `fn (?*?*T, U) void`: if the on_load function can never fail.\n- `fn (?*?*T, U) !void`: if the on_load function can fail with an error.\n  The module load integer will reflect the integer value of the error.\n  > #### Zig error integers {: .warning }\n  >\n  > Note that the integer representation of a zig error may change between\n  > compilations.  Translating this integer back to a meaningful value may\n  > be challenging.\n- `fn (?*?*T, U) int`: the on_load function will be considered to fail if\n  the integer value is not `0`.\n- `fn (?*?*T, U) E`: for an `enum` type `E`, the on_load function will be\n  considered to fail if the enum value is not zero.  It's recommended that\n  the enum type `E` be defined as so: `const E = enum{ ok = 0, ...};`\n- `fn (beam.env, ?*?*anyopaque, e.ErlNifTerm) c_int`: this is a \n  \"raw\" call that corresponds to the expected signature of a `upgrade` callback.","ref":"10-callbacks.html#on_load"},{"type":"extras","title":"on_load Callback types - Module Callbacks","doc":"`T` may be any type at all, which is considered the *private data* of \nthe module.  Any information may be stored in a `*T`, and the double \npointer is passed to the callback for the pointer to be communicated \nto the VM.  The `*T` supplied will be accessible via the `enif_priv_data`\nfunction\n\n`U` must be a type that can be passed as the first argument of `beam.make`, \nand the value of `U` will be set by the result of an `__on_load__/0` function \ndefined in the module body.  The result of this function call will be passed\nto the `on_load`\n\n>","ref":"10-callbacks.html#on_load-callback-types"},{"type":"extras","title":"on_load and resources {: .info } - Module Callbacks","doc":">\n> The beam nif guide says that resources must be initialized in the `load`\n> callback.  The `on_load` callback must NOT initialize resources.  This\n> is performed in a function that will wrap your `on_load` callback.\n\n#### Example\n\n```elixir\ndefmodule OnLoadExample do\n  use ExUnit.Case, async: true\n  use Zig, otp_app: :zigler, callbacks: [on_load: :load_fn]\n\n  defp __on_load__, do: 47\n\n  ~Z\"\"\"\n  const beam = @import(\"beam\");\n  const e = @import(\"erl_nif\");\n\n  pub fn load_fn(private: ?*?*u32, number: u32) !void {\n      const stored_pointer = try beam.allocator.create(u32);\n      stored_pointer.* = number;\n      private.?.* = stored_pointer;\n  }\n\n  pub fn get_private() u32 {\n      const priv_ptr: ?*anyopaque = e.enif_priv_data(beam.context.env);\n      const priv_ptr_u32: *u32 = @ptrCast(@alignCast(priv_ptr.?));\n      return priv_ptr_u32.*;\n  }\n  \"\"\"\n\n  test \"on_load stores value into private\" do\n    assert 47 = get_private()\n  end\nend\n#module\n```","ref":"10-callbacks.html#on_load-and-resources-info"},{"type":"extras","title":"on_upgrade - Module Callbacks","doc":"The on_upgrade callback may have one of the following function signatures:\n\n- `fn (?*?*T, ?*?*U, V) void`: if the on_load function can never fail.\n- `fn (?*?*T, ?*?*U, V) !void`: if the on_load function can fail with an error.\n  The module load integer will reflect the integer value of the error.\n  > #### Zig error integers {: .warning }\n  >\n  > Note that the integer representation of a zig error may change between\n  > compilations.  Translating this integer back to a meaningful value may\n  > be challenging.\n- `fn (?*?*T, ?*?*U, V) int`: the on_load function will be considered to fail if\n  the integer value is not `0`.\n- `fn (?*?*T, ?*?*U, V) E`: for an `enum` type `E`, the on_load function will be\n  considered to fail if the enum value is not zero.  It's recommended that\n  the enum type `E` be defined as so: `const E = enum{ ok = 0, ...};`\n- `fn (beam.env, ?*?*anyopaque, ?*?*anyopaque, e.ErlNifTerm) c_int`: this is a \n  \"raw\" call that corresponds to the expected signature of a `upgrade` callback.","ref":"10-callbacks.html#on_upgrade"},{"type":"extras","title":"on_upgrade Callback types - Module Callbacks","doc":"`T` and `U` may be any type at all, which is considered the *private data* of \nthe module.  Any information may be stored in these pointers, and the double \npointer is passed to the callback for the pointer to be communicated \nto the VM.  The `*U` data supplied will be accessible via the `enif_priv_data`\nfunction\n\n`V` must be a type that can be passed as the first argument of `beam.make`, \nand the value of `V` will be set by the result of an `__on_load__/0` function \ndefined in the module body.  The result of this function call will be passed\nto the `on_load`\n\n>","ref":"10-callbacks.html#on_upgrade-callback-types"},{"type":"extras","title":"on_upgrade and resources {: .info } - Module Callbacks","doc":">\n> The beam nif guide says that resources must be initialized in the `upgrade`\n> callback.  The `on_upgrade` callback must NOT initialize resources.  This\n> is performed in a function that will wrap your `on_upgrade` callback.","ref":"10-callbacks.html#on_upgrade-and-resources-info"},{"type":"extras","title":"on_unload - Module Callbacks","doc":"The on_unload callback may have one of the following function signatures:\n\n- `fn (?*T) void`: if you would like your inbound private data to be typed.\n- `fn (beam.env, ?*anyopaque) void`: this is a \"raw\" call that corresponds \n  to the expected signature of a `upgrade` callback.","ref":"10-callbacks.html#on_unload"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,50]}}